/*
Universidade Federal do Rio Grande do Norte - UFRN
Federal University of Rio Grande do Norte

Authors:
Arthur D. L. Lima (arthurlima@dca.ufrn.br)
Luiz F. Q. Silveira (lfelipe@dca.ufrn.br)
Samuel Xavier-de-Souza (samuel@dca.ufrn.br)

Description: Parallel Algorithm for Cyclostationary Feature Extraction

This parallel algorithm is based on the following serial version:
Z. Zhang, X. Xu, Implementation of cyclic periodogram detection on VEE for
cognitive radio, in: Global Mobile Congress, Mobile World Congress Shangai,
2007, pp. 1-5.
URL: https://www.keysight.com/upload/cmc_upload/All/5989_7480EN.pdf

To compile:
gcc pcpd.c -o pcpd -fopenmp -lm -std=c99 -lfftw3f_omp -lfftw3f

export OMP_NUM_THREADS=1
*/

#include <stdio.h>
#include <stdlib.h>
#include <complex.h>
#include <math.h>
#include <sys/time.h>
#include <omp.h>
#include <fftw3.h>
#include <time.h>
// Load the input signal from a file
#include "load_signal.c"
// Save S in S_re.txt and S_im.txt
#include "save_signal.c"
#include "mylib.c"

int main(int argc, char *argv[]){

	// ./pcpd N L n_loops
	if(argc<4){
		printf("./pcpd N L n_loops\n");
		exit(-1);
	}

	// blocks size;
	const int N = atoi( argv[1] );
	const int L = atoi( argv[2] );
	// total number of samples
	const int len = N*L;
	// number of iterations for Monte Carlo simulations
	const int n_loops = atoi( argv[3] );
	float t_history[n_loops];
	printf("./pcpd N=%d L=%d n_loops=%d\n", N, L, n_loops);

	FILE *file;
	file = fopen("../results/results.txt","a+");

	// measure elapsed time for iterations
	struct timeval t_begin, t_end;
	gettimeofday(&t_begin, NULL);

	// sampling rate
	const int Fs = 2048;
	// cyclic frequency resolution
	const int dalpha = 2;
	// df = M*dalpha, frequency resolution
	const int df = 32*dalpha;
	// smoothing factor in frequency axis
	const int M = floor(df/dalpha);
	//const int N = Fs/dalpha;
	//const int L = floor(len/N); // how many blocks
	const int T1_len = (N*N+N)/2;
	const int LM = L*M; // averaging by 1/(L*M)in the Step 5
	const int newk = floor(N/M); // smoothing in S
	// Max numbers of threads in the parallel region
	const int max_threads = omp_get_max_threads();
	const int n_procs = omp_get_num_procs(); //Number os processors avalilable

	// To analyze a signal from a file, uncomment the following two lines
	//int len2 = 0;
	//float _Complex* x = load_signal(&len2); // input signal
	// To analyze a signal from a file, comment the following lines
	float _Complex* x = gen_rand_sig( len );

	// Dynamically alocate arrays xl, Xl, Tl1,  T1 and S
	float _Complex **xl = alloc1d_p(L); // Step 1, *xl[L]
	float _Complex **Xl = alloc2d(L, N); // Step 2, Xl[L][N]
	float _Complex **Tl1 = alloc2d(L, T1_len); // Step 3, Tl1[L][(N*N+N)/2]
	float _Complex *T1 = alloc1d(T1_len); // Step 4, T1[(N*N+N)/2]
	float _Complex **S = alloc2d(N, newk); // Step 5, S[N][newk];

	printf("Parameters: Fs=%d dalpha=%d df=%d M=%d N=%d L=%d len=%d\n",
		Fs, dalpha, df, M, N, L, len);
	printf("Max number of threads: %d/%d\n", max_threads, n_procs);

	// This function should be called before any other FFTW's
	if( fftwf_init_threads() == 0 ){
		fprintf(stderr,"Error on initialization of the FFTW threads\n");
		exit(-1);
	}

	// This function should be called before creating a plan we want to
	//paralelize
	//fftwf_plan_with_nthreads( max_threads );
	fftwf_plan_with_nthreads( 1 );
	fftwf_plan p[L];

	// Plan creation is not thread safe
	// also do the Step 1
	for(int l=0; l<L; l++){
		// Step 1 - Divides input sample sequence (x) into L blocks, with
		// each block containing N samples (xl)
		xl[l] =  x + (l * N) ;
		// Plan creation
		p[l] = fftwf_plan_dft_1d(
			N, xl[l], Xl[l], FFTW_FORWARD, FFTW_ESTIMATE);
	}

	struct timeval tmin_begin, tmin_end;
	float t_min = 100000.0;
	float tmin_cur;

for(int z=0; z<n_loops; z++)
{
	gettimeofday(&tmin_begin, NULL);
	#pragma omp parallel default(none)  \
		shared( x, xl, Xl, S, p, T1, Tl1)
		//Loop index variables are by default private
	{
		// Step 2 - Takes discrete Fourier transform of the N samples
		//in each block l
		#pragma omp for schedule( dynamic )
		for(int l=0; l<L; l++){
			fftwf_execute( p[l] ); // thread safe
		}


		// Step 3 - Calculates
		// Tl[k][alpha][l] = Xl[l][k] * conj( Xl[l][k-alpha] ) / N
		// k=0..N-1, l=0..L-1, alpha = 0..N-1
		#pragma omp for schedule(guided) collapse(2) nowait
		for(int l=0; l<L; l++){
			for(int k=0; k<N; k++){
				for(int alpha=0; alpha<(N-k); alpha++){
						//int offset = ( alpha * (2*N - alpha + 1) ) / 2;;
						int offset = ( alpha * ((N<<1) - alpha + 1) ) >> 1;;
						Tl1[l][offset+k] =  Xl[l][k] * conj( Xl[l][k+alpha] ) / N ;
				}
			}

		}

		#pragma omp for collapse(2) schedule( dynamic ) nowait
		for(int i=0; i<N; i++){
			for(int k=0; k<newk; k++){
				S[i][k] = 0;
			}
		}

		#pragma omp for schedule( guided )
		for(int i=0; i<T1_len; i++){
			T1[i] = 0;
		}

		//Step 4 - Averages the results of step 3 over all blocks
		for(int l=0; l<L; l++){
			#pragma omp for
			for(int k=0; k<T1_len; k++){
				T1[k] += Tl1[l][k];
			}
		}

		//Step 5 - Smoothes in frequency domain to obtain the cyclic periodogram
		//Guided influence negativelly the performance
		#pragma omp for collapse(2) schedule ( guided )
		for(int alpha=0; alpha<N; alpha++){
			//int start = alpha/2;
			//int end = N - start - (alpha%2) - 1;
			//int Tl_offset = ( alpha * (2*N - alpha + 1) ) / 2;

			for(int k=0; k<newk; k++){
				for(int m=0; m<M; m++){
					int start = alpha>>1;
					int end = N - start - (alpha&1) - 1;
					int Tl_offset = ( alpha * ( (N<<1) - alpha + 1) ) >> 1;
					int k_aux = M*k+m;
					if( (k_aux>=start) && (k_aux<=end)){
						S[alpha][k] += T1[Tl_offset + k_aux - start];
					}
				}
				S[alpha][k] /= LM;
			}
		}


	} // End of parallel region

	gettimeofday(&tmin_end, NULL);
	tmin_cur = (tmin_end.tv_sec + tmin_end.tv_usec / 1e6) -
		(tmin_begin.tv_sec + tmin_begin.tv_usec / 1e6);
	t_history[z] = tmin_cur;
	if(tmin_cur < t_min)
		t_min = tmin_cur;
} // End of the loop to get the least time

	// Save the variable S, the SCD
	save_signal(S, N, newk);

	// Free x, xl, Xl, Tl1, T1 and S
	free(x); x = NULL;
	free(xl); xl = NULL;
	free2d(Xl, L);
	free2d(Tl1, L);
	free(T1); T1=NULL;
	free2d(S, N);

	// plan destruction is not thread safe
	for(int l=0; l<L; l++){
		fftwf_destroy_plan(p[l]);
	}
	fftwf_cleanup();
	fftwf_cleanup_threads();

	// Sort the array
	qsort(t_history, n_loops, sizeof(float), float_cmp) ;
	float median = t_history[n_loops/2];

	gettimeofday(&t_end, NULL);
	float t_total = (t_end.tv_sec + t_end.tv_usec / 1e6) -
		(t_begin.tv_sec + t_begin.tv_usec / 1e6);

	printf("Min.Time: %f Median: %f Total time: %f Threads: %d/%d N: %d L: %d n_loops: %d\n",
		t_min, median, t_total, max_threads, n_procs, N, L, n_loops);
	//fprintf(file, "Min.Time: %f Median: %f Total time: %f Threads: %d/%d N: %d L: %d n_loops: %d\n",
	//	t_min, median, t_total, max_threads, n_procs, N, L, n_loops);
	fprintf(file, "%f,%d,%d,%d,%d,%d\n", median, max_threads, n_procs, N, L, n_loops);
  fclose(file);
	return 0;
}
